# Written by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: AptListener.py 123 2007-06-23 20:48:11Z camrdale-guest $

"""Listen for download requests from Apt.

@type alas: C{string}
@var alas: the message to send when the data is not found
@type VERSION: C{string}
@var VERSION: the Server identifier sent to all sites

"""

from DebTorrent.parseargs import parseargs, formatDefinitions
from DebTorrent.RawServer import RawServer, autodetect_ipv6, autodetect_socket_style
from DebTorrent.HTTPHandler import HTTPHandler, months, weekdays
from DebTorrent.parsedir import parsedir
from DebTorrent.subnetparse import IP_List, ipv6_to_ipv4, to_ipv4, is_valid_ip, is_ipv4
from DebTorrent.iprangeparse import IP_List as IP_Range_List
from DebTorrent.torrentlistparse import parsetorrentlist
from threading import Event, Thread
from DebTorrent.bencode import bencode, bdecode, Bencached
from DebTorrent.zurllib import urlopen, quote, unquote
from Filter import Filter
from urlparse import urlparse
from os import rename, getpid
from os.path import exists, isfile, join
from cStringIO import StringIO
from gzip import GzipFile
from bz2 import decompress
from traceback import print_exc
from time import time, gmtime, strftime, localtime
from DebTorrent.clock import clock
from random import shuffle, seed, randrange
from sha import sha
from types import StringType, IntType, LongType, ListType, DictType
from binascii import b2a_hex, a2b_hex, a2b_base64
from string import lower
from makemetafile import getpieces, getsubpieces, uniconvert, convert_all
from DebTorrent.HTTPCache import HTTPCache
import sys, os
import signal
import re
import DebTorrent.__init__
from DebTorrent.__init__ import version, createPeerID, product_name,version_short

try:
    True
except:
    True = 1
    False = 0
    bool = lambda x: not not x
    
DEBUG = True

VERSION = product_name+'/'+version_short

def statefiletemplate(x):
    """Check the saved state file for corruption.
    
    @type x: C{dictionary}
    @param x: the dictionary of information retrieved from the state file
    @raise ValueError: if the state file info is corrupt
    
    """
    
    if type(x) != DictType:
        raise ValueError
    for cname, cinfo in x.items():
        if cname == 'peers':
            for y in cinfo.values():      # The 'peers' key is a dictionary of SHA hashes (torrent ids)
                if type(y) != DictType:   # ... for the active torrents, and each is a dictionary
                    raise ValueError
                for id, info in y.items(): # ... of client ids interested in that torrent
                    if (len(id) != 20):
                        raise ValueError
                    if type(info) != DictType:  # ... each of which is also a dictionary
                        raise ValueError # ... which has an IP, a Port, and a Bytes Left count for that client for that torrent
                    if type(info.get('ip', '')) != StringType:
                        raise ValueError
                    port = info.get('port')
                    if type(port) not in (IntType,LongType) or port < 0:
                        raise ValueError
                    left = info.get('left')
                    if type(left) not in (IntType,LongType) or left < 0:
                        raise ValueError
                    if type(info.get('supportcrypto')) not in (IntType,LongType):
                        raise ValueError
                    if type(info.get('requirecrypto')) not in (IntType,LongType):
                        raise ValueError
        elif cname == 'completed':
            if (type(cinfo) != DictType): # The 'completed' key is a dictionary of SHA hashes (torrent ids)
                raise ValueError          # ... for keeping track of the total completions per torrent
            for y in cinfo.values():      # ... each torrent has an integer value
                if type(y) not in (IntType,LongType):
                    raise ValueError      # ... for the number of reported completions for that torrent
        elif cname == 'allowed':
            if (type(cinfo) != DictType): # a list of info_hashes and included data
                raise ValueError
            if x.has_key('allowed_dir_files'):
                adlist = [z[1] for z in x['allowed_dir_files'].values()]
                for y in cinfo.keys():        # and each should have a corresponding key here
                    if not y in adlist:
                        raise ValueError
        elif cname == 'allowed_dir_files':
            if (type(cinfo) != DictType): # a list of files, their attributes and info hashes
                raise ValueError
            dirkeys = {}
            for y in cinfo.values():      # each entry should have a corresponding info_hash
                if not y[1]:
                    continue
                if not x['allowed'].has_key(y[1]):
                    raise ValueError
                if dirkeys.has_key(y[1]): # and each should have a unique info_hash
                    raise ValueError
                dirkeys[y[1]] = 1
            

alas = 'your file may exist elsewhere in the universe\nbut alas, not here\n'


def isotime(secs = None):
    """Create an ISO formatted string of the time.
    
    @type secs: C{float}
    @param secs: number of seconds since the epoch 
        (optional, default is to use the current time)
    @rtype: C{string}
    @return: the ISO formatted string representation of the time
    
    """
    
    if secs == None:
        secs = time()
    return strftime('%Y-%m-%d %H:%M UTC', gmtime(secs))

class AptListener:
    """Listen for Apt requests to download files.
    
    @type handler: unknown
    @ivar handler: the download handler to use
    @type config: C{dictionary}
    @ivar config: the configuration parameters
    @type dfile: C{string}
    @ivar dfile: the state file to use when saving the current state
    @type parse_dir_interval: C{int}
    @ivar parse_dir_interval: seconds between reloading of the allowed 
        directory or file, and the lists of allowed and banned IPs
    @type favicon: C{string}
    @ivar favicon: file containing x-icon data
    @type rawserver: L{DebTorrent.RawServer}
    @ivar rawserver: the server to use for scheduling
    @type times: unknown
    @ivar times: unknown
    @type state: C{dictionary}
    @ivar state: the current state information for the tracking
    @type allowed_IPs: unknown
    @ivar allowed_IPs: unknown
    @type banned_IPs: unknown
    @ivar banned_IPs: unknown
    @type allowed_ip_mtime: unknown
    @ivar allowed_ip_mtime: unknown
    @type banned_ip_mtime: unknown
    @ivar banned_ip_mtime: unknown
    @type trackerid: unknown
    @ivar trackerid: unknown
    @type save_dfile_interval: C{int}
    @ivar save_dfile_interval: seconds between saving the state file
    @type show_names: C{boolean}
    @ivar show_names: whether to display names from allowed dir
    @type prevtime: unknown
    @ivar prevtime: unknown
    @type logfile: unknown
    @ivar logfile: unknown
    @type log: unknown
    @ivar log: unknown
    @type allow_get: unknown
    @ivar allow_get: unknown
    @type allowed: unknown
    @ivar allowed: unknown
    @type allowed_list_mtime: unknown
    @ivar allowed_list_mtime: unknown
    @type allowed_dir_files: unknown
    @ivar allowed_dir_files: unknown
    @type allowed_dir_blocked: unknown
    @ivar allowed_dir_blocked: unknown
    @type uq_broken: unknown
    @ivar uq_broken: unknown
    @type Filter: unknown
    @ivar Filter: unknown
    @type Cache: L{DebTorrent.HTTPCache.HTTPCache}
    @ivar Cache: the cache of downloaded files
    @type cache_waiting: C{dictionary}
    @ivar cache_waiting: the pending HTTP get requests that are waiting for download 
        from the cache. Keys are strings that are the path being requested, values
        are lists of L{DebTorrent.HTTPHandler.HTTPConnection} objects which are the
        requests that are pending for that path.
    @type request_queue: C{dictionary}
    @ivar request_queue: the pending HTTP get requests that are waiting for download.
        Keys are L{DebTorrent.HTTPHandler.HTTPConnection} objects, values are
        (L{DebTorrent.download_bt1.BT1Download}, C{int}, C{list} of C{int}, C{float})
        which are the torrent downloader, file index, list of pieces needed, and 
        the time of the original request.
    
    """

    def __init__(self, handler, config, rawserver):
        """Initialize the instance.
        
        @type handler: unknown
        @param handler: the download handler to use
        @type config: C{dictionary}
        @param config: the configuration parameters
        @type rawserver: L{DebTorrent.RawServer}
        @param rawserver: the server to use for scheduling
        
        """

        self.handler = handler
        self.config = config
        self.dfile = config['dfile']
        favicon = config['favicon']
        self.parse_dir_interval = config['apt_parse_dir_interval']
        self.favicon = None
        if favicon:
            try:
                h = open(favicon,'r')
                self.favicon = h.read()
                h.close()
            except:
                print "**warning** specified favicon file -- %s -- does not exist." % favicon
        self.rawserver = rawserver
        self.times = {}
        self.state = {}

        self.allowed_IPs = None
        self.banned_IPs = None
        if config['allowed_ips'] or config['banned_ips']:
            self.allowed_ip_mtime = 0
            self.banned_ip_mtime = 0
            self.read_ip_lists()
                
        if exists(self.dfile):
            try:
                h = open(self.dfile, 'rb')
                ds = h.read()
                h.close()
                tempstate = bdecode(ds)
                if not tempstate.has_key('peers'):
                    tempstate = {'peers': tempstate}
                statefiletemplate(tempstate)
                self.state = tempstate
            except:
                print '**warning** statefile '+self.dfile+' corrupt; resetting'

        self.trackerid = createPeerID('-T-')
        seed(self.trackerid)
                
        self.save_dfile_interval = config['save_dfile_interval']
        self.show_names = config['show_names']
        #rawserver.add_task(self.save_state, self.save_dfile_interval)
        self.prevtime = clock()
        self.logfile = None
        self.log = None
        if (config['logfile']) and (config['logfile'] != '-'):
            try:
                self.logfile = config['logfile']
                self.log = open(self.logfile,'a')
                sys.stdout = self.log
                print "# Log Started: ", isotime()
            except:
                print "**warning** could not redirect stdout to log file: ", sys.exc_info()[0]

        if config['hupmonitor']:
            def huphandler(signum, frame, self = self):
                try:
                    self.log.close ()
                    self.log = open(self.logfile,'a')
                    sys.stdout = self.log
                    print "# Log reopened: ", isotime()
                except:
                    print "**warning** could not reopen logfile"
             
            signal.signal(signal.SIGHUP, huphandler)            
                
        self.allow_get = config['allow_get']
        
        if config['allowed_list']:
            if config['allowed_dir']:
                print '**warning** allowed_dir and allowed_list options cannot be used together'
                print '**warning** disregarding allowed_dir'
                config['allowed_dir'] = ''
            self.allowed = self.state.setdefault('allowed_list',{})
            self.allowed_list_mtime = 0
            self.parse_allowed()
            self.remove_from_state('allowed','allowed_dir_files')

        elif config['allowed_dir']:
            self.allowed = self.state.setdefault('allowed',{})
            self.allowed_dir_files = self.state.setdefault('allowed_dir_files',{})
            self.allowed_dir_blocked = {}
            self.parse_allowed()
            self.remove_from_state('allowed_list')

        else:
            self.allowed = None
            self.remove_from_state('allowed','allowed_dir_files', 'allowed_list')
                
        self.uq_broken = unquote('+') != ' '
        self.Filter = Filter(rawserver.add_task)
        self.Cache = HTTPCache(rawserver)
        self.cache_waiting = {}
        
        self.request_queue = {}
        rawserver.add_task(self.process_queue, 1)
        
    def enqueue_request(self, connection, downloader, file_num, pieces_needed):
        """Add a new download request to the queue of those waiting for pieces.
        
        @type connection: L{DebTorrent.HTTPHandler.HTTPConnection}
        @param connection: the conection the request came in on
        @type downloader: L{DebTorrent.download_bt1.BT1Download}
        @param downloader: the torrent download that has the file
        @type file_num: C{int}
        @param file_num: the index of the file in the torrent
        @type pieces_needed: C{list} of C{int}
        @param pieces_needed: the list of pieces in the torrent that still 
            need to download
        
        """
        
        assert not self.request_queue.has_key(connection)
        
        if DEBUG:
            print 'queueing request as file', file_num, 'needs pieces:', pieces_needed

        self.request_queue[connection] = (downloader, file_num, pieces_needed, clock())
        
    def process_queue(self):
        """Process the queue of waiting requests."""
        
        # Schedule it again
        self.rawserver.add_task(self.process_queue, 1)
        
        closed_conns = []
        for c, v in self.request_queue.items():
            # Check for a closed connection
            if c.closed:
                closed_conns.append(c)
                continue
                
            # Remove the downloaded pieces from the list of needed ones
            for piece in list(v[2]):
                if v[0].storagewrapper.do_I_have(piece):
                    if DEBUG:
                        print 'queued request for file', v[1], 'got piece', piece
                    v[2].remove(piece)
                    
            # If no more pieces are needed, return the answer and remove the request
            if not v[2]:
                if DEBUG:
                    print 'queued request for file', v[1], 'is complete'
                del self.request_queue[c]
                self.answer_package(c, v[0], v[1])

        # Remove closed connections from the queue
        for c in closed_conns:
            del self.request_queue[c]


    def get_infopage(self):
        """Format the info page to display for normal browsers.
        
        Formats the currently downloading torrents into a table in human-readable
        format to display in a browser window.
        
        @rtype: (C{int}, C{string}, C{dictionary}, C{string})
        @return: the HTTP status code, status message, headers, and message body
        
        """
        
        try:
            if not self.config['show_infopage']:
                return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)
            red = self.config['infopage_redirect']
            if red:
                return (302, 'Found', {'Server': VERSION, 'Content-Type': 'text/html', 'Location': red},
                        '<A HREF="'+red+'">Click Here</A>')
            
            # Write the document header
            s = StringIO()
            s.write('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" ' \
                    '"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' \
                    '<html><head><title>DebTorrent download info</title>\n')
            if self.favicon is not None:
                s.write('<link rel="shortcut icon" href="/favicon.ico">\n')
            s.write('</head>\n<body>\n' \
                    '<h3>DebTorrent download info</h3>\n'\
                    '<ul>\n'
                    '<li><strong>client version:</strong> %s</li>\n' \
                    '<li><strong>client time:</strong> %s</li>\n' \
                    '</ul>\n' % (version, isotime()))
            
            # Write the table headers
            s.write('<table summary="files" border="1">\n' \
                    '<tr><th>name</th>\n' \
                    '<th>status</th>\n' \
                    '<th align="right">progress</th>\n' \
                    '<th align="right">peers</th>\n' \
                    '<th align="right">seeds</th>\n' \
                    '<th align="right">distributed copies</th>\n' \
                    '<th align="right">upload rate</th>\n' \
                    '<th align="right">download rate</th>\n' \
                    '<th align="right">uploaded</th>\n' \
                    '<th align="right">downloaded</th>\n' \
                    '<th align="right">size</th>\n' \
                    '<th align="right">time remaining</th>\n' \
                    '<th>error message</th></tr>\n')

            # Get the data from the statistics gatherer
            data = self.handler.gather_stats()
            if not data:
                s.write('<tr><td colspan="13" align="left">no torrents</td></tr>\n')
                
            # Display a table row for each running torrent
            for x in data:
                ( name, status, progress, peers, seeds, seedsmsg, dist,
                  uprate, dnrate, upamt, dnamt, size, t, msg ) = x
                s.write('<tr><td>%s</td>\n' \
                        '<td>%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%.3f</td>\n' \
                        '<td align="right">%0.1fK/s</td>\n' \
                        '<td align="right">%0.1fK/s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td align="right">%s</td>\n' \
                        '<td>%s</td></tr>\n' \
                        % (name, status, progress, peers, seeds, dist, 
                           uprate/1000, dnrate/1000, size_format(upamt), 
                           size_format(dnamt), size_format(size), hours(t), msg))

            s.write('</table>\n')

            s.write('</body>\n' \
                '</html>\n')
            return (200, 'OK', {'Server': VERSION, 'Content-Type': 'text/html; charset=iso-8859-1'}, s.getvalue())
        except:
            print_exc()
            return (500, 'Internal Server Error', {'Server': VERSION, 'Content-Type': 'text/html; charset=iso-8859-1'}, 'Server Error')


    def get_meow(self):
        return (200, 'OK', {'Server': VERSION, 'Content-Type': 'text/html; charset=iso-8859-1'}, """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n<html><head><title>Meow</title>\n</head>\n<body style="color: rgb(255, 255, 255); background-color: rgb(0, 0, 0);">\n<div><big style="font-weight: bold;"><big><big><span style="font-family: arial,helvetica,sans-serif;">I&nbsp;IZ&nbsp;TAKIN&nbsp;BRAKE</span></big></big></big><br></div>\n<pre><b><tt>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; .-o=o-.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ,&nbsp; /=o=o=o=\ .--.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; _|\|=o=O=o=O=|&nbsp;&nbsp;&nbsp; \<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; __.'&nbsp; a`\=o=o=o=(`\&nbsp;&nbsp; /<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; '.&nbsp;&nbsp; a 4/`|.-""'`\ \ ;'`)&nbsp;&nbsp; .---.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; \&nbsp;&nbsp; .'&nbsp; /&nbsp;&nbsp; .--'&nbsp; |_.'&nbsp;&nbsp; / .-._)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; `)&nbsp; _.'&nbsp;&nbsp; /&nbsp;&nbsp;&nbsp;&nbsp; /`-.__.' /<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; `'-.____;&nbsp;&nbsp;&nbsp;&nbsp; /'-.___.-'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; `\"""`</tt></b></pre>\n<div><big style="font-weight: bold;"><big><big><span style="font-family: arial,helvetica,sans-serif;">FRM&nbsp;GETIN&nbsp;UR&nbsp;PACKAGES</span></big></big></big><br></div>\n</body>\n</html>""")


    def get_cached(self, connection, path, headers):
        """Proxy the (possibly cached) download of a file from a mirror.
        
        @type connection: L{DebTorrent.HTTPHandler.HTTPConnection}
        @param connection: the conection the request came in on
        @type path: C{list} of C{string}
        @param path: the path of the file to download, starting with the mirror name
        @type headers: C{dictionary}
        @param headers: the headers from the request
        @rtype: (C{int}, C{string}, C{dictionary}, C{string})
        @return: the HTTP status code, status message, headers, and downloaded file
            (or None if the file is being downloaded)
        
        """
        
        try:
            # Deb files don't need to be checked for staleness
            uptodate = True
            if path[-1][-4:] == '.deb':
                uptodate = False

            # First check the cache for the file
            r = self.Cache.cache_get(path, uptodate, headers.get('if-modified-since', ''))
            
            # If the cache doesn't have it
            if r[0] not in (200, 304):
                # Get Debs from the debtorrent download, others are straight download
                if path[-1][-4:] == '.deb':
                    return self.get_package(connection, path)
                else:
                    # Save the connection info and start downloading the file
                    self.cache_waiting.setdefault('/'.join(path), []).append(connection)
                    self.Cache.download_get(path, self.get_cached_callback)
                    return None
            
            if path[-1] in ('Packages', 'Packages.gz', 'Packages.bz2'):
                if r[0] == 304:
                    # Oops, we do need the cached file after all to start the torrent
                    r2 = self.Cache.cache_get(path)
                    self.got_Packages(path, r2[3])
                else:
                    self.got_Packages(path, r[3])                    

            return r
        
        except IOError, e:
            try:
                (msg, status) = e
            except:
                status = 404
                msg = 'Unknown error occurred'
            return (status, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, msg)
            
    def get_cached_callback(self, path, r):
        """Return the newly cached file to the waiting connection.
        
        @type path: C{list} of C{string}
        @param path: the path of the file to download, starting with the mirror name
        @type r: (C{int}, C{string}, C{dictionary}, C{string})
        @param r: the HTTP status code, status message, headers, and cached data
        
        """

        # Get the list of connections waiting for this file
        connections = self.cache_waiting.pop('/'.join(path), None)
        
        if connections is None:
            if DEBUG:
                print 'no connection exists to return the cached file on'
            return

        # If it's a torrent file, start it
        if r[0] == 200 and path[-1] in ('Packages', 'Packages.gz', 'Packages.bz2'):
            self.got_Packages(path, r[3])

        for connection in connections:
            # Check to make sure the requester is still waiting
            if connection.closed:
                continue
            
            connection.answer(r)
            
    def get_package(self, connection, path):
        """Download a package file from a torrent.
        
        @type connection: L{DebTorrent.HTTPHandler.HTTPConnection}
        @param connection: the conection the request came in on
        @type path: C{list} of C{string}
        @param path: the path of the file to download, starting with the mirror name
        @rtype: (C{int}, C{string}, C{dictionary}, C{string})
        @return: the HTTP status code, status message, headers, and package data
            (or None if the package is to be downloaded)
        
        """

        # Find the file in one of the torrent downloads
        d, f = self.handler.find_file(path[0], path[1:])
        
        if d is None:
            return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)

        # Check if the file has already been downloaded
        data = ''
        pieces_needed = []
        start_piece, end_piece = d.fileselector.storage.file_pieces[f]
        for piece in xrange(start_piece, end_piece+1):
            if not d.storagewrapper.do_I_have(piece):
                pieces_needed.append(piece)
            elif not pieces_needed:
                data = data + d.storagewrapper.get_piece(piece, 0, -1).getarray().tostring()
        
        if not pieces_needed:
            return (200, 'OK', {'Server': VERSION, 'Content-Type': 'text/plain'}, data)

        # Check if the torrent is running/not paused
        if d.doneflag.isSet():
            return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)
        
        if not d.unpauseflag.isSet():
            d.Unpause()
        
        # Enable the download of the piece
        d.fileselector.set_priority(f, 1)
        
        # Add the connection to the list of those needing responses
        self.enqueue_request(connection, d, f, pieces_needed)
        
        return None
        
    
    def answer_package(self, connection, d, f):
        """Send the newly downloaded package file to the requester.
        
        @type connection: L{DebTorrent.HTTPHandler.HTTPConnection}
        @param connection: the conection the request came in on
        @type d: L{DebTorrent.download_bt1.BT1Download}
        @param d: the torrent download that has the file
        @type f: C{int}
        @param f: the index of the file in the torrent
        
        """

        # Check to make sure the requester is still waiting
        if connection.closed:
            return

        # Check if the file has been downloaded
        data = ''
        pieces_needed = []
        start_piece, end_piece = d.fileselector.storage.file_pieces[f]
        for piece in xrange(start_piece, end_piece+1):
            if not d.storagewrapper.do_I_have(piece):
                pieces_needed.append(piece)
            elif not pieces_needed:
                data = data + d.storagewrapper.get_piece(piece, 0, -1).getarray().tostring()
        
        if not pieces_needed:
            connection.answer((200, 'OK', {'Server': VERSION, 'Content-Type': 'text/plain'}, data))
            return

        # Something strange has happened, requeue it
        if DEBUG:
            print 'request for', f, 'still needs pieces:', pieces_needed
        self.enqueue_request(connection, d, f, pieces_needed)
        
    
    def got_Packages(self, path, data):
        """Process a downloaded Packages file and start a torrent.
        
        @type path: C{list} of C{string}
        @param path: the path of the file to download, starting with the mirror name
        @type data: C{string}
        @param data: the downloaded Packages file
        
        """

        try:
            # Decompress the data
            if path[-1].endswith('.gz'):
                compressed = StringIO(data)
                f = GzipFile(fileobj = compressed)
                data = f.read()
            elif path[-1].endswith('.bz2'):
                data = decompress(data)
            
            name = '_'.join(path[:-1])

            assert data[:8] == "Package:"
            h = data.split('\n')
        except:
            if DEBUG:
                print 'ERROR: Packages file could not be converted to a torrent'
            return 

        sub_pieces = getsubpieces('_'.join(path))

        (info, info_all) = getpieces(h, separate_all = self.config['separate_all'], sub_pieces = sub_pieces)
        
        if info and self.config['separate_all'] in (0, 2, 3):
            self.start_torrent(info, name, path)
            
        if info_all and self.config['separate_all'] in (1, 3):
            self.start_torrent(info_all, convert_all(name), path)

    def start_torrent(self, info, name, path):
        response = {'info': info,
                    'announce': self.config['default_tracker'], 
                    'name': uniconvert(name)}

        if path.count('dists'):
            mirror = 'http://' + '/'.join(path[:path.index('dists')]) + '/'
            response['deb_mirrors'] = [mirror]
        
        infohash = sha(bencode(response['info'])).digest()
        
        a = {}
        a['path'] = '/'.join(path)
        a['file'] = name
        a['type'] = path[-1]
        mirror = path[0]
        for i in path[1:path.index('dists')]:
            mirror = join(mirror, i)
        a['mirror'] = mirror
        i = response['info']
        l = 0
        nf = 0
        if i.has_key('length'):
            l = i.get('length',0)
            nf = 1
        elif i.has_key('files'):
            for li in i['files']:
                nf += 1
                if li.has_key('length'):
                    l += li['length']
        a['numfiles'] = nf
        a['length'] = l
        a['name'] = name
        def setkey(k, d = response, a = a):
            if d.has_key(k):
                a[k] = d[k]
        setkey('failure reason')
        setkey('warning message')
        setkey('announce-list')
        a['metainfo'] = response
        
        self.handler.add(infohash, a)
        
    def get(self, connection, path, headers):
        """Respond to a GET request.
        
        Process a GET request from APT/browser/other. Process the request,
        calling the helper functions above if needed. Return the response to
        be returned to the requester.
        
        @type connection: L{DebTorrent.HTTPHandler.HTTPConnection}
        @param connection: the conection the request came in on
        @type path: C{string}
        @param path: the URL being requested
        @type headers: C{dictionary}
        @param headers: the headers from the request
        @rtype: (C{int}, C{string}, C{dictionary}, C{string})
        @return: the HTTP status code, status message, headers, and message body
        
        """
        
#        return (404, 'Not Found', {'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)
        real_ip = connection.get_ip()
        ip = real_ip
        if is_ipv4(ip):
            ipv4 = True
        else:
            try:
                ip = ipv6_to_ipv4(ip)
                ipv4 = True
            except ValueError:
                ipv4 = False

        if ( (self.allowed_IPs and not self.allowed_IPs.includes(ip))
             or (self.banned_IPs and self.banned_IPs.includes(ip)) ):
            return (400, 'Not Authorized', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'},
                bencode({'failure reason':
                'your IP is not allowed on this proxy'}))

        paramslist = {}
        def params(key, default = None, l = paramslist):
            """Get the user parameter, or the default.
            
            @type key: C{string}
            @param key: the parameter to get
            @type default: C{string}
            @param default: the default value to use if no parameter is set
                (optional, defaults to None)
            @type l: C{dictionary}
            @param l: the user parameters (optional, defaults to L{paramslist})
            @rtype: C{string}
            @return: the parameter's value
            
            """
            
            if l.has_key(key):
                return l[key][0]
            return default

        try:
            (scheme, netloc, path, pars, query, fragment) = urlparse(path)
            if self.uq_broken == 1:
                #path = path.replace('+',' ') What is this!
                query = query.replace('+',' ')
            path = unquote(path)[1:]
            for s in query.split('&'):
                if s:
                    i = s.index('=')
                    kw = unquote(s[:i])
                    paramslist.setdefault(kw, [])
                    paramslist[kw] += [unquote(s[i+1:])]
                    
            if path == '' or path == 'index.html':
                return self.get_infopage()
            if path == 'meow':
                return self.get_meow()
            if path == 'favicon.ico':
                if self.favicon is not None:
                    return (200, 'OK', {'Server': VERSION, 'Content-Type' : 'image/x-icon'}, self.favicon)
                else:
                    return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)

            # Process the rest as a proxy
            path = path.split('/')

            if 'Packages.diff' in path:
                return (404, 'Not Found', {'Server': VERSION, 'Content-Type': 'text/plain', 'Pragma': 'no-cache'}, alas)
            
            return self.get_cached(connection, path, headers)
            
        except ValueError, e:
            return (400, 'Bad Request', {'Server': VERSION, 'Content-Type': 'text/plain'}, 
                'you sent me garbage - ' + str(e))


    def remove_from_state(self, *l):
        """Remove all the input parameter names from the current state."""
        for s in l:
            try:
                del self.state[s]
            except:
                pass

    def save_state(self):
        """Save the state file to disk."""
        self.rawserver.add_task(self.save_state, self.save_dfile_interval)
        h = open(self.dfile, 'wb')
        h.write(bencode(self.state))
        h.close()


    def parse_allowed(self):
        """Periodically parse the directory and list for allowed torrents."""
        self.rawserver.add_task(self.parse_allowed, self.parse_dir_interval)

        if self.config['allowed_dir']:
            r = parsedir( self.config['allowed_dir'], self.allowed,
                          self.allowed_dir_files, self.allowed_dir_blocked,
                          [".dtorrent"] )
            ( self.allowed, self.allowed_dir_files, self.allowed_dir_blocked,
                added, garbage2 ) = r
            
            self.state['allowed'] = self.allowed
            self.state['allowed_dir_files'] = self.allowed_dir_files

            self.t2tlist.parse(self.allowed)
            
        else:
            f = self.config['allowed_list']
            if self.allowed_list_mtime == os.path.getmtime(f):
                return
            try:
                r = parsetorrentlist(f, self.allowed)
                (self.allowed, added, garbage2) = r
                self.state['allowed_list'] = self.allowed
            except (IOError, OSError):
                print '**warning** unable to read allowed torrent list'
                return
            self.allowed_list_mtime = os.path.getmtime(f)

        for infohash in added.keys():
            self.downloads.setdefault(infohash, {})
            self.completed.setdefault(infohash, 0)
            self.seedcount.setdefault(infohash, 0)


    def read_ip_lists(self):
        """Periodically parse the allowed and banned IPs lists."""
        self.rawserver.add_task(self.read_ip_lists,self.parse_dir_interval)
            
        f = self.config['allowed_ips']
        if f and self.allowed_ip_mtime != os.path.getmtime(f):
            self.allowed_IPs = IP_List()
            try:
                self.allowed_IPs.read_fieldlist(f)
                self.allowed_ip_mtime = os.path.getmtime(f)
            except (IOError, OSError):
                print '**warning** unable to read allowed_IP list'
                
        f = self.config['banned_ips']
        if f and self.banned_ip_mtime != os.path.getmtime(f):
            self.banned_IPs = IP_Range_List()
            try:
                self.banned_IPs.read_rangelist(f)
                self.banned_ip_mtime = os.path.getmtime(f)
            except (IOError, OSError):
                print '**warning** unable to read banned_IP list'
                

def size_format(s):
    """Format a byte size for reading by the user.
    
    @type s: C{long}
    @param s: the number of bytes
    @rtype: C{string}
    @return: the formatted size with appropriate units
    
    """
    
    if (s < 1024):
        r = str(s) + 'B'
    elif (s < 1048576):
        r = str(int(s/1024)) + 'KiB'
    elif (s < 1073741824L):
        r = str(int(s/1048576)) + 'MiB'
    elif (s < 1099511627776L):
        r = str(int((s/1073741824.0)*100.0)/100.0) + 'GiB'
    else:
        r = str(int((s/1099511627776.0)*100.0)/100.0) + 'TiB'
    return(r)

def hours(n):
    """Formats seconds into a human-readable time.
    
    Formats a given number of seconds into a human-readable time appropriate
    for display to the user.
    
    @type n: C{int}
    @param n: the number of seconds
    @rtype: C{string}
    @return: a displayable representation of the number of seconds
    
    """
    
    if n == 0:
        return 'complete!'
    try:
        n = int(n)
        assert n >= 0 and n < 5184000  # 60 days
    except:
        return '<unknown>'
    m, s = divmod(n, 60)
    h, m = divmod(m, 60)
    if h > 0:
        return '%d hour %02d min %02d sec' % (h, m, s)
    else:
        return '%d min %02d sec' % (m, s)

